(define-derived-mode emop-mode fundamental-mode "EMoP"
  "Major mode to edit EMoP files."
  (set
   (make-local-variable 'font-lock-defaults)
   '(emop-font-lock-keywords
     nil nil nil
     emop-beginning-of-clause))
  (set
   (make-local-variable 'beginning-of-defun-function)
   'emop-beginning-of-clause)
  (set (make-local-variable 'indent-line-function) 'emop-indent-line)
  (set (make-local-variable 'comment-start) "/* ")
  (set (make-local-variable 'comment-end) " */")
  (set (make-local-variable 'comment-column) 40)
  (set (make-local-variable 'comment-start-skip) "/\\*+[ \t]*")
  (set (make-local-variable 'comment-multi-line) t)
  (set (make-local-variable 'indent-tabs-mode) nil)
  (set (make-local-variable 'emop-indent-width) 3)
  (set-syntax-table emop-mode-syntax-table)
  (if (fboundp 'run-mode-hooks)
      (run-mode-hooks 'emop-mode-hook)
    (run-hooks 'emop-mode-hook)))

(defvar emop-mode-syntax-table
  (let ((table (make-syntax-table)))
    (modify-syntax-entry ?\" "_" table)
    (modify-syntax-entry ?_ "w" table)
    (modify-syntax-entry ?\\ "\\" table)
    (modify-syntax-entry ?/ ". 14" table)
    (modify-syntax-entry ?* ". 23" table)
    (modify-syntax-entry ?+ "." table)
    (modify-syntax-entry ?- "." table)
    (modify-syntax-entry ?= "." table)
    (modify-syntax-entry ?# "<" table)
    (modify-syntax-entry ?\n ">" table)
    (modify-syntax-entry ?< "." table)
    (modify-syntax-entry ?> "." table)
    table))

(defun emop-beginning-of-clause (&optional arg)
  "Move backward to the beginning of a clause."
  (interactive "p")
  (unless arg (setq arg 1))
  (while (> arg 1)
    (beginning-of-line)
    (if (re-search-backward "[.}]$" 0 t)
        (if (looking-at "}")
            (if (not (re-searck-backward "{" 0 t))
                (setq arg 0))
          (setq arg (1- arg)))
      (setq arg 0)))
  (if (re-search-backward "[.{]$" 0 t)
      (emop-next-code-line)
    (beginning-of-buffer)))

(defun emop-next-code-line ()
  "Go to the next code line (skipping empty lines and comments)."
  (interactive)
  (while
      (and
       (not (eobp))
       (forward-line 1)
       (looking-at "[[:blank:]]*\\(%\\|/\\*\\|$\\)"))
    (if (looking-at "[[:blank:]]*/\\*")
        (while
            (and
             (not (looking-at "[^\n]*\\*/"))
             (not (eobp))
             (forward-line 1))
          t))))

(defvar emop-escape-sequence
  "\\([\\]\\([nrtb'\"\\]\\|x[0-9A-Fa-f][0-9A-Fa-f]?\\|[0-9]\\([0-9][0-9]?\\)?\\)\\)")

(defvar emop-font-lock-keywords
  (list
   '("/\\*\\*\\([^*]\\|\\*[^/]\\)*\\*/"
     0 font-lock-doc-face)

   '("/\\*\\([^*]\\|\\*[^/]\\)*\\*/\\|%[^\n]*\n"
     0 font-lock-comment-face)

   (list
    (concat
     "\\(\\(^\\|[^0]\\)\\('\\([^\\']\\|"
     emop-escape-sequence
     "\\|[\\]\n\\)*'\\)\\|\\(\\)\\(\\)\\(\"\\([^\"\\]\\|"
     emop-escape-sequence
     "\\|[\\]\n\\)*\"\\)\\)")
    3 font-lock-string-face)

   '("\\<\\(module\\)\\>"
     0 font-lock-keyword-face)

   '("\\(\\<\\(mod\\|quo\\)\\>\\|[][(){}]\\|\\.\\|:-\\|;\\|->\\||\\|,\\|\\\\[+]\\|<=>\\|=>\\|\\\\/\\|/\\\\\\|\\\\=\\|=\\|<\\|>\\|=<\\|>=\\|\\.\\.\\|[+]\\|-\\|[*]\\|/\\|[*][*]\\|[+][+]\\|[-][-]\\|:\\)"
     0 font-lock-builtin-face)

   (list
    (concat
     "\\<\\(0\\(x[0-9A-Fa-f]+\\|b[0-1]+\\|'\\([^\\]\\|"
     emop-escape-sequence
     "\\)\\|o?[0-7]+\\|\\)\\|[1-9][0-9]*\\(e\\(-\\)?[0-9]+\\)?\\)\\>")
    0 font-lock-constant-face)

   '("\\(\\<[a-z][a-zA-Z0-9_]*\\>\\)\\((\\([^.:]\\|:[^-]\\)*)\\)? :-"
     1 font-lock-function-name-face)

   '("\\<[A-Z_][a-zA-Z0-9_']*\\>"
     0 font-lock-variable-name-face))

  "Font-Lock keywords for EMoP mode.")

(defun emop-indent-line ()
  "Indent current line as EMoP code."
  (interactive "p")
  (let ((indent (emop-indent-level))
  (pos (- (point-max) (point))))
    (beginning-of-line)
    (indent-line-to indent)
    (if (> (- (point-max) pos) (point))
  (goto-char (- (point-max) pos)))))

(defun emop-indent-level ()
  "Compute EMoP indentation level."
  (save-excursion
    (let ((less nil))
      (beginning-of-line)
      (skip-chars-forward " \t")
      (if (looking-at "\\([])};]\\|->\\)")
          (setq less t))
      (beginning-of-line)
      (if (bobp)
          0
        (forward-line -1)
        (while
            (and
             (not (eobp))
             (looking-at "[[:blank:]]*\n"))
          (forward-line -1))
        (if (looking-at "[[:blank:]]*\n")
            0
          (skip-chars-forward " \t")
          (let ((indent (current-column)))
            (if (looking-at "[^\n]*\\.[[:blank:]]*\n")
                (let ((line (line-number-at-pos nil)))
                  (emop-beginning-of-clause)
                  (if (/= line (line-number-at-pos nil))
                      (setq indent (- indent emop-indent-width))))
              (if (looking-at "[^\n%]*\\(:-\\|=\\|{\\|\\[\\|(\\|->\\|;\\)[[:blank:]]*\n")
                  (setq indent (+ indent emop-indent-width))))
            (if less
                (setq indent (- indent emop-indent-width)))
            indent))))))
